"""
This module is built on top of the abstraction

It adds the ability to have multiple arities and names attached to functions.

Because MyPy doesn't support multiple arity generics, we need a seperate class for each arity of
function.
"""
from __future__ import annotations


from metadsl import *
from .rules import *
from .abstraction import *
import typing

__all__ = ["FunctionZero", "FunctionOne", "FunctionTwo", "FunctionThree"]


T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")
X = typing.TypeVar("X")


class FunctionZero(Expression, typing.Generic[T]):
    """
    Function with zero args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[], T]) -> FunctionZero[T]:
        return cls.create(fn.__name__, fn())

    @expression
    def to_fn(self) -> typing.Callable[[], T]:
        ...

    @expression
    def __call__(self) -> T:
        ...

    @expression
    @classmethod
    def create(cls, name: str, val: T) -> FunctionZero[T]:
        ...

    @expression
    def value(self) -> T:
        ...


class FunctionOne(Expression, typing.Generic[T, U]):
    """
    Function with one arg.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T], U]) -> FunctionOne[T, U]:
        return cls.create(fn.__name__, Abstraction.from_fn(fn))

    @expression
    def to_fn(self) -> typing.Callable[[T], U]:
        ...

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionOne[T, U], T], U]
    ) -> FunctionOne[T, U]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(inner_abst: Abstraction[T, U]) -> Abstraction[T, U]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> U:
                return fn(inner_fn, arg1)

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg: T) -> U:
        ...

    @expression
    @classmethod
    def create(cls, name: str, val: Abstraction[T, U]) -> FunctionOne[T, U]:
        ...

    @expression
    def abstraction(self) -> Abstraction[T, U]:
        ...


class FunctionTwo(Expression, typing.Generic[T, U, V]):
    """
    Function with two args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T, U], V]) -> FunctionTwo[T, U, V]:
        def inner(arg1: T) -> Abstraction[U, V]:
            def inner(arg2: U) -> V:
                return fn(arg1, arg2)

            return Abstraction.from_fn(inner)

        return cls.create(fn.__name__, Abstraction.from_fn(inner))

    @expression
    def to_fn(self) -> typing.Callable[[T, U], V]:
        ...

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionTwo[T, U, V], T, U], V]
    ) -> FunctionTwo[T, U, V]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(
            inner_abst: Abstraction[T, Abstraction[U, V]]
        ) -> Abstraction[T, Abstraction[U, V]]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> Abstraction[U, V]:
                @Abstraction.from_fn
                def inner(arg2: U) -> V:
                    return fn(inner_fn, arg1, arg2)

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg1: T, arg2: U) -> V:
        ...

    @expression
    @classmethod
    def create(
        cls, name: str, val: Abstraction[T, Abstraction[U, V]]
    ) -> FunctionTwo[T, U, V]:
        ...

    @expression
    def abstraction(self) -> Abstraction[T, Abstraction[U, V]]:
        ...


class FunctionThree(Expression, typing.Generic[T, U, V, X]):
    """
    Function with three args.
    """

    @expression
    @classmethod
    def from_fn(cls, fn: typing.Callable[[T, U, V], X]) -> FunctionThree[T, U, V, X]:
        @Abstraction.from_fn
        def inner(arg1: T) -> Abstraction[U, Abstraction[V, X]]:
            @Abstraction.from_fn
            def inner(arg2: U) -> Abstraction[V, X]:
                @Abstraction.from_fn
                def inner(arg3: V) -> X:
                    return fn(arg1, arg2, arg3)

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def to_fn(self) -> typing.Callable[[T, U, V], X]:
        ...

    @expression
    @classmethod
    def from_fn_recursive(
        cls, fn: typing.Callable[[FunctionThree[T, U, V, X], T, U, V], X]
    ) -> FunctionThree[T, U, V, X]:
        @Abstraction.fix
        @Abstraction.from_fn
        def inner(
            inner_abst: Abstraction[T, Abstraction[U, Abstraction[V, X]]]
        ) -> Abstraction[T, Abstraction[U, Abstraction[V, X]]]:
            inner_fn = cls.create(fn.__name__, inner_abst)

            @Abstraction.from_fn
            def inner(arg1: T) -> Abstraction[U, Abstraction[V, X]]:
                @Abstraction.from_fn
                def inner(arg2: U) -> Abstraction[V, X]:
                    @Abstraction.from_fn
                    def inner(arg3: V) -> X:
                        return fn(inner_fn, arg1, arg2, arg3)

                    return inner

                return inner

            return inner

        return cls.create(fn.__name__, inner)

    @expression
    def __call__(self, arg1: T, arg2: U, arg3: V) -> X:
        ...

    @expression
    @classmethod
    def create(
        cls, name: str, val: Abstraction[T, Abstraction[U, Abstraction[V, X]]]
    ) -> FunctionThree[T, U, V, X]:
        ...

    @expression
    def abstraction(self) -> Abstraction[T, Abstraction[U, Abstraction[V, X]]]:
        ...


register_pre(default_rule(FunctionZero[T].from_fn))
register_pre(default_rule(FunctionOne[T, U].from_fn))
register_pre(default_rule(FunctionTwo[T, U, V].from_fn))
register_pre(default_rule(FunctionThree[T, U, V, X].from_fn))

register_pre(default_rule(FunctionOne[T, U].from_fn_recursive))
register_pre(default_rule(FunctionTwo[T, U, V].from_fn_recursive))
register_pre(default_rule(FunctionThree[T, U, V, X].from_fn_recursive))


@register  # type: ignore
@rule
def zero_call(fn: FunctionZero[T]) -> R[T]:
    return fn(), fn.value()


@register  # type: ignore
@rule
def one_call(fn: FunctionOne[T, U], t: T) -> R[U]:
    return fn(t), fn.abstraction()(t)


@register  # type: ignore
@rule
def two_call(fn: FunctionTwo[T, U, V], t: T, u: U) -> R[V]:
    return fn(t, u), fn.abstraction()(t)(u)


@register  # type: ignore
@rule
def three_call(fn: FunctionThree[T, U, V, X], t: T, u: U, v: V) -> R[X]:
    return fn(t, u, v), fn.abstraction()(t)(u)(v)


@register  # type: ignore
@rule
def zero_value(_: str, t: T) -> R[T]:
    return FunctionZero.create(_, t).value(), t


@register  # type: ignore
@rule
def one_abstraction(_: str, abst: Abstraction[T, U]) -> R[Abstraction[T, U]]:
    return FunctionOne.create(_, abst).abstraction(), abst


@register  # type: ignore
@rule
def two_abstraction(
    _: str, abst: Abstraction[T, Abstraction[U, V]]
) -> R[Abstraction[T, Abstraction[U, V]]]:
    return FunctionTwo.create(_, abst).abstraction(), abst


@register  # type: ignore
@rule
def three_abstraction(
    _: str, abst: Abstraction[T, Abstraction[U, Abstraction[V, X]]]
) -> R[Abstraction[T, Abstraction[U, Abstraction[V, X]]]]:
    return FunctionThree.create(_, abst).abstraction(), abst



# from __future__ import annotations
# import numbers

# from metadsl import *
# import typing

# __all__ = ["Integer", "Tuple", "Number", "Optional", "Union"]

# T = typing.TypeVar("T")
# U = typing.TypeVar("U")


# class Boolean(Expression):
#     @staticmethod
#     @expression
#     def from_bool(b: bool) -> Boolean:
#         ...


# class Integer(Expression):
#     @staticmethod
#     @expression
#     def from_int(i: int) -> Integer:
#         ...

#     @expression
#     def __add__(self, other: Integer) -> Integer:
#         ...

#     @expression
#     def __mul__(self, other: Integer) -> Integer:
#         ...


# class Tuple(Expression, typing.Generic[T]):
#     """
#     Mirrors the Python Tuple API, of a homogenous type.
#     """

#     @expression
#     def __getitem__(self, index: Integer) -> T:
#         ...

#     @classmethod
#     @expression
#     def from_items(cls, *items: T) -> Tuple[T]:
#         ...

#     @expression
#     def rest(self) -> Tuple[T]:
#         ...


# class Number(Expression):
#     """
#     Floating point, integer, or complex number.
#     """

#     @staticmethod
#     @expression
#     def from_number(i: numbers.Number) -> Number:
#         ...

#     @expression
#     def __add__(self, other: Number) -> Number:
#         ...

#     @expression
#     def __mul__(self, other: Number) -> Number:
#         ...




# """

# This is a replica of the NumPy API that can be used without requiring NumPy.

# You should be able to import from `metadsl.nonumpy.compat` just as you would from `numpy` and use all
# the functions and objects you are accustomed to.

# What will it actually do? Well not much. It will just build up an expression tree of the NumPy expression you have written.

# It also comes included with some transformations of those expression trees you can use. It's a good building block if you are writing a NumPy-like API.

# Things it does not preserve:

# 1. Primitive Python operations like `__str__` and `type`
# 2. Control flow operations like `if` and `for`
# 3. The C API
# 4. Mutation operations (like setting a slice)

# It should be just as accepting as the default NumPy API in the type of objects it takes.
# """
# import typing
# import metadsl.numpy.wraps
# import dataclasses

# __all__ = ["arange"]

# @dataclasses.dataclass
# class NDArray:
#     array: metadsl.numpy.wraps.NDArray

#     @property
#     def shape(self) -> Tuple[Integer]:
#         return self.array.shape()

#     def __add__(self, other: typing.Any) -> NDArray:
#         return self.array + other

#     @expression
#     def __mul__(self, other: NDArray) -> NDArray:
#         ...

#     @expression
#     def __getitem__(self, idxs: Union[Integer, Tuple[Integer]]) -> NDArray:
#         ...


# If arange is called with one positional argument, it is the stop, and you cannot set the start with a kward
# @typing.overload
# def arange(stop, *, step=None, dtype=None) -> NDArray:
#     ...


# # If arange is called with two positional arguments, the first is stop, and the second is start
# @typing.overload
# def arange(start, stop, step=None, dtype=None) -> NDArray:
#     ...


# def arange(*args, **kwargs) -> NDArray:
#     start: typing.Any
#     if len(args) == 1:
#         stop, = args
#         start = None
#         assert kwargs.keys() < {"step", "dtype"}
#         step, dtype = kwargs.get("step"), kwargs.get("dtype")
#     elif len(args) == 2:
#         start, stop = args
#         assert kwargs.keys() < {"step", "dtype"}
#         step, dtype = kwargs.get("step"), kwargs.get("dtype")
#     elif len(args) == 3:
#         start, stop, step = args
#         assert kwargs.keys() < {"dtype"}
#         dtype = kwargs.get("dtype")
#     elif len(args) == 4:
#         start, stop, step, dtype = args
#         assert not kwargs.keys()
#     else:
#         raise RuntimeError("Cannot pass more than 4 arguments to `arange`")
#     return NDArray.from_pure(
#         np_pure.arange(
#             create_instance(
#                 instance_type(py_pure.Optional, instance_type(py_pure.Number)), start
#             ),
#             create_instance(instance_type(py_pure.Number), stop),
#             create_instance(
#                 instance_type(py_pure.Optional, instance_type(py_pure.Number)), step
#             ),
#             create_instance(
#                 instance_type(py_pure.Optional, instance_type(np_pure.DType)), dtype
#             ),
#         )
#     )


#     # def __gt__(self, other) -> "NDArray":
#     #     return self.from_pure(
#     #         self.pure > create_instance(instance_type(np_pure.NDArray), other)
#     #     )

#     # def __ge__(self, other) -> "NDArray":
#     #     return self.from_pure(
#     #         self.pure >= create_instance(instance_type(np_pure.NDArray), other)
#     #     )

#     # def __lt__(self, other) -> "NDArray":
#     #     return self.from_pure(
#     #         self.pure < create_instance(instance_type(np_pure.NDArray), other)
#     #     )

#     # def __le__(self, other) -> "NDArray":
#     #     return self.from_pure(
#     #         self.pure <= create_instance(instance_type(np_pure.NDArray), other)
#     #     )

#     # def __array_ufunc__(self, ufunc, method, *inputs, **kwargs) -> "NDArray":
#     #     return self.from_pure(
#     #         np_pure.NDArray.__array_ufunc__(
#     #             create_instance(instance_type(np_pure.UFunc), ufunc),
#     #             create_instance(instance_type(np_pure.UFuncMethod), method),
#     #             create_instance(instance_type(np_pure.UFuncKwargs), kwargs),
#     #             *(
#     #                 create_instance(instance_type(np_pure.NDArray), input)
#     #                 for input in inputs
#     #             )
#     #         )
#     #     )

#     # def __array_function__(self, func, types, args, kwargs) -> "NDArray":
#     #     return self.from_pure(
#     #         np_pure.NDArray.__array_function__(
#     #             create_instance(instance_type(np_pure.NumPyFunction), func),
#     #             create_instance(
#     #                 instance_type(py_pure.Tuple, instance_type(Instance)), args
#     #             ),
#     #             create_instance(instance_type(Instance), kwargs),
#     #         )
#     #     )


# # class UFunc(Instance):
# #     @classmethod
# #     def from_pure(cls, p: np_pure.UFunc) -> "UFunc":
# #         return cls(p.__value__)

# #     @property
# #     def pure(self) -> np_pure.UFunc:
# #         return np_pure.UFunc(self.__value__)


# # class UFuncKwargs(Instance):
# #     @classmethod
# #     def from_pure(cls, p: np_pure.UFuncKwargs) -> "UFuncKwargs":
# #         return cls(p.__value__)

# #     @property
# #     def pure(self) -> np_pure.UFuncKwargs:
# #         return np_pure.UFuncKwargs(self.__value__)


# # class UFuncMethod(Instance):
# #     @classmethod
# #     def from_pure(cls, p: np_pure.UFuncMethod) -> "UFuncMethod":
# #         return cls(p.__value__)

# #     @property
# #     def pure(self) -> np_pure.UFuncMethod:
# #         return np_pure.UFuncMethod(self.__value__)


# # def array(object, dtype=None, copy=True, order=None, subok=None, ndmin=None) -> NDArray:
# #     return NDArray.from_pure(
# #         np_pure.array(
# #             create_instance(instance_type(np_pure.NDArray), object),
# #             create_instance(
# #                 instance_type(py_pure.Optional, instance_type(np_pure.DType)), dtype
# #             ),
# #             create_instance(instance_type(py_pure.Boolean), copy),
# #             create_instance(
# #                 instance_type(py_pure.Optional, instance_type(np_pure.ArrayOrder)),
# #                 order,
# #             ),
# #             create_instance(
# #                 instance_type(py_pure.Optional, instance_type(py_pure.Boolean)), subok
# #             ),
# #             create_instance(
# #                 instance_type(py_pure.Optional, instance_type(py_pure.Integer)), ndmin
# #             ),
# #         )
# #     )




# from __future__ import annotations

# import typing

# from metadsl import *
# import metadsl.numpy.expressions as np
# import metadsl.python.wraps as py_wraps

# __all__ = ["arange", "NDArray"]

# T = typing.TypeVar("T")


# @wrap(np.arange)
# def arange(start, stop, step, dtype) -> NDArray:
#     ...


# class NDArray(Wrap[np.NDArray]):
#     @wrap(np.NDArray.shape)
#     def shape(self) -> py_wraps.IntegerTuple:
#         ...

#     @wrap(np.NDArray.__add__)
#     def __add__(self, other: object) -> NDArray:
#         ...

#     @wrap(np.NDArray.__mul__)
#     def __mul__(self, other: object) -> NDArray:
#         ...

#     @wrap(np.NDArray.__getitem__)
#     def __getitem__(self, idxs: object) -> NDArray:
#         ...



"""
Normalized expressions, for deduping and single replacements.

IDs should be consistant accross replacements.
"""
from __future__ import annotations
import dataclasses
import typing
import itertools
import collections
import functools


from .expressions import *


__all__ = [
    "ExpressionReference",
    "NormalizedExpression",
    "NormalizedExpressions",
    "Parent",
    "Children",
    "hash_value",
    "Hash",
    "ID",
]

ID = typing.NewType("ID", int)
Hash = typing.NewType("Hash", int)


@dataclasses.dataclass(frozen=True)
class Parent:
    hash: Hash
    key: typing.Union[str, int]


@functools.singledispatch
def hash_value(value: object) -> int:
    """
    Computes some hash for a value that should be stable. Either use the built in hash, or if we cannot
    (like the object is mutable) then use the id.

    It's a single dispatch function so that you can register custom hashes for objects you don't control.
    """
    try:
        return Hash(hash((type(value), value)))
    except TypeError:
        return Hash(hash((type(value), id(value))))


@dataclasses.dataclass
class Children:
    args: typing.List[Hash]
    kwargs: typing.Dict[str, Hash]

    def references(self) -> typing.Iterable[typing.Tuple[typing.Union[str, int], Hash]]:
        return itertools.chain(enumerate(self.args), self.kwargs.items())


def compute_hash(value: object, children: typing.Optional[Children]) -> Hash:
    if children:
        assert isinstance(value, Expression)
        return Hash(
            hash(
                (
                    value.function,
                    tuple(children.args),
                    frozenset(children.kwargs.items()),
                )
            )
        )
    return Hash(hash_value(value))


@dataclasses.dataclass
class NormalizedExpression:
    children: typing.Optional[Children]
    value: object
    id: ID
    parents: typing.Set[Parent] = dataclasses.field(default_factory=set)


@dataclasses.dataclass
class NormalizedExpressions:
    # Expressions, topologically sorted, with the root expression at the end.

    expressions: collections.OrderedDict[
        Hash, NormalizedExpression
    ] = dataclasses.field(default_factory=collections.OrderedDict)
    ids: typing.Dict[ID, Hash] = dataclasses.field(default_factory=dict)

    next_id: ID = ID(0)
    # THe first hash  added. Used when adding a new expression, so we can record the left
    # most hash and delete  every hash before that
    _first_hash: typing.Optional[Hash] = None

    def add(
        self,
        expr: object,
        prev: typing.Optional[
            typing.Tuple[
                typing.Optional[Hash],
                collections.OrderedDict[Hash, NormalizedExpression],
            ]
        ] = None,
    ) -> Hash:
        """
        Does a depth addition of the expression to the graph, keeping all the nodes
        added in topological order, because the children will all be visisted after
        the parent.
        """

        children: typing.Optional[Children]

        prev_id: typing.Optional[ID]
        prev_args: typing.Optional[typing.List[Hash]]
        prev_kwargs: typing.Optional[typing.Dict[str, Hash]]
        prev_expressions: typing.Optional[
            collections.OrderedDict[Hash, NormalizedExpression]
        ]

        if prev:
            prev_hash, prev_expressions = prev
            if prev_hash:
                prev_ref = prev_expressions[prev_hash]
                prev_id = prev_ref.id
                if (
                    isinstance(prev_ref.value, Expression)
                    and isinstance(expr, Expression)
                    and prev_ref.value.function == expr.function
                ):
                    assert prev_ref.children
                    prev_args = prev_ref.children.args
                    prev_kwargs = prev_ref.children.kwargs
                else:
                    prev_args = None
                    prev_kwargs = None
            else:
                prev_id = None
                prev_args = None
                prev_kwargs = None
        else:
            prev_expressions = None
            prev_id = None
            prev_args = None
            prev_kwargs = None

        if isinstance(expr, Expression):
            children = Children([], {})
            for i, arg in enumerate(expr.args):
                arg_hash = self.add(
                    arg,
                    (
                        prev_args[i] if prev_args and i < len(prev_args) else None,
                        prev_expressions,
                    )
                    if prev_expressions
                    else None,
                )
                children.args.append(arg_hash)

                # Update expression with child, so that points to same one
                expr.args[i] = self.expressions[arg_hash].value
            for k, v in expr.kwargs.items():
                kwarg_hash = self.add(
                    v,
                    (
                        prev_kwargs.get(k, None) if prev_kwargs else None,
                        prev_expressions,
                    )
                    if prev_expressions
                    else None,
                )
                children.kwargs[k] = kwarg_hash

                # Update expression with child, so that points to same one
                expr.kwargs[k] = self.expressions[kwarg_hash].value
        else:
            children = None

        hash = compute_hash(expr, children)

        if children:
            for key, child_hash in children.references():
                self.expressions[child_hash].parents.add(Parent(hash, key))

        if not self._first_hash:
            self._first_hash = hash
        if hash not in self.expressions:
            id = self._get_new_id(prev_id)
            self.expressions[hash] = NormalizedExpression(children, expr, id)
            self.ids[id] = hash
        return hash

    def replace(self, hash: typing.Optional[Hash], expr: object) -> None:
        new_expressions = NormalizedExpressions(next_id=self.next_id)
        if not hash:
            # If we didn't get a hash, we are replacing the root node
            root_hash = next(iter(reversed(self.expressions)))
            root_expr = expr
        else:
            # Otherwise, we are replacing a child node
            # In which case, we update its parents, and get the root noode
            for parent in self.expressions[hash].parents:
                value = self.expressions[parent.hash].value
                assert isinstance(value, Expression)
                if isinstance(parent.key, str):
                    value.kwargs[parent.key] = expr
                else:
                    value.args[parent.key] = expr

            root_hash = self._get_root(hash)
            root_expr = self.expressions[root_hash].value
        new_expressions.add(root_expr, (root_hash, self.expressions))
        self.expressions = new_expressions.expressions
        self.ids = new_expressions.ids
        self.next_id = new_expressions.next_id

    def _get_new_id(self, prev_id: typing.Optional[ID]) -> ID:
        if prev_id and prev_id not in self.ids:
            return prev_id
        id = self.next_id
        self.next_id = ID(self.next_id + 1)
        assert id not in self.ids
        return id

    def _get_root(self, hash: Hash) -> Hash:
        parents = list(self.expressions[hash].parents)
        if parents:
            return self._get_root(parents[0].hash)
        return hash

    def _assert_expressions_topological_sorted(self) -> None:
        """
        All expressions should be after their children
        """
        for i, expr in enumerate(self.expressions.values()):
            if not expr.children:
                continue
            for _, child_hash in expr.children.references():
                # Children should be to the left of their parents
                assert self._index_of_hash(child_hash) < i

    def _index_of_hash(self, hash: Hash) -> int:

        for i, hash_ in enumerate(self.expressions.keys()):
            if hash_ == hash:
                return i
        raise RuntimeError()

    def _all_children(self, hash: Hash) -> typing.Set[Hash]:
        ref = self.expressions[hash]
        s = {hash}
        if not ref.children:
            return s
        for _, child_hash in ref.children.references():
            s.update(self._all_children(child_hash))
        return s

    def _assert_parents_children_consistant(self) -> None:
        """
        Asserts that parents match the children.

        For each expression, verifies that all children have this node as a parent, and all parents
        have this node as a child.
        """

        for hash, ref in self.expressions.items():
            for parent in ref.parents:
                key = parent.key
                parent_ref = self.expressions[parent.hash]
                assert parent_ref.children
                if isinstance(key, str):
                    assert parent_ref.children.kwargs[key] == hash
                else:
                    assert parent_ref.children.args[key] == hash

            if not ref.children:
                continue
            for key, child_hash in ref.children.references():
                child_ref = self.expressions[child_hash]
                assert Parent(key=key, hash=hash) in child_ref.parents

    def _assert_hashes_accurate(self):
        for hash, ref in self.expressions.items():
            assert hash == compute_hash(ref.value, ref.children)

    def _assert_children_match_values(self):
        for hash, ref in self.expressions.items():
            value = ref.value
            if not ref.children:
                assert not isinstance(value, Expression)
                continue
            assert isinstance(value, Expression)

            for h, arg in zip(ref.children.args, value.args):
                assert self.expressions[h].value is arg

            for k, v in ref.children.kwargs.items():
                assert value.kwargs[k] is self.expressions[k].value

    def _assert_ids_match(self):
        for id, hash in self.ids.items():
            assert self.expressions[hash].id == id


T = typing.TypeVar("T")


@dataclasses.dataclass
class ExpressionReference(typing.Generic[T]):
    # Hash that we refer to. If it is None, then it is the last  expression by default
    expressions: NormalizedExpressions
    _hash: typing.Optional[Hash] = dataclasses.field(default=None, repr=False)

    def __post_init__(self):
        self.verify_integrity()

    @property
    def hash(self) -> Hash:
        return self._hash or self.root_hash

    @property
    def root_hash(self) -> Hash:
        return next(iter(reversed(self.expressions.expressions)))

    @classmethod
    def from_expression(cls, expr: T) -> ExpressionReference:
        expressions = NormalizedExpressions()
        expressions.add(expr)
        return cls(expressions)

    @property
    def normalized_expression(self) -> NormalizedExpression:
        return self.expressions.expressions[self.hash]

    def replace(self, new_expr: T) -> None:
        self.expressions.replace(self._hash, new_expr)
        self._hash = None

        # Optional, this could be disabled to improve performance
        self.verify_integrity()

    @property
    def children(self) -> typing.Iterable[ExpressionReference]:
        """
        Returns all the children of the graph, starting from the root node
        and ending in the leaf nodes.
        """
        root_node = True
        for k, v in reversed(self.expressions.expressions.items()):
            yield ExpressionReference(self.expressions, None if root_node else k)
            root_node = False

    def verify_integrity(self) -> None:
        """
        Verifies a number of properties about the data  structures
        """

        expressions = self.expressions.expressions

        assert self.hash in expressions

        # Order
        self.expressions._assert_expressions_topological_sorted()

        all_children = self.expressions._all_children(self.root_hash)
        all_keys = set(expressions.keys())
        assert all_children == all_keys

        # IDs
        assert len(self.expressions.ids) == len(expressions)
        self.expressions._assert_ids_match()

        # Parents
        self.expressions._assert_parents_children_consistant()

        # Hashes
        self.expressions._assert_hashes_accurate()

        # Children
        self.expressions._assert_children_match_values()



from __future__ import annotations

import typing

__all__ = [
    # "Module",
    # "RegisteredFunctionOne",
    # "RegisteredFunctionTwo",
    # "RegisteredFunctionThree",
]
from metadsl import *
from metadsl_core import *

T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")
X = typing.TypeVar("X")


class Module(Expression):
    @expression
    @classmethod
    def create(cls, name: str) -> Module:
        ...

    @expression
    def register_function_one(
        self, fn: FunctionOne[T, U]
    ) -> Pair[Module, RegisteredFunctionOne[T, U]]:
        ...

    @expression
    def register_function_two(
        self, fn: FunctionTwo[T, U, V]
    ) -> Pair[Module, RegisteredFunctionTwo[T, U, V]]:
        ...

    @expression
    def register_function_three(
        self, fn: FunctionThree[T, U, V, X]
    ) -> Pair[Module, RegisteredFunctionThree[T, U, V, X]]:
        ...


class RegisteredFunctionOne(Expression, typing.Generic[T, U]):
    @expression
    def __call__(self, a: T) -> U:
        ...

    @expression
    def to_fn(self) -> typing.Callable[[T], U]:
        ...


class RegisteredFunctionTwo(Expression, typing.Generic[T, U, V]):
    @expression
    def __call__(self, a: T, b: U) -> V:
        ...

    @expression
    def to_fn(self) -> typing.Callable[[T, U], V]:
        ...


class RegisteredFunctionThree(Expression, typing.Generic[T, U, V, X]):
    @expression
    def __call__(self, a: T, b: U, c: V) -> X:
        ...

    @expression
    def to_fn(self) -> typing.Callable[[T, U, V], X]:
        ...



from __future__ import annotations
import typing
import ctypes


from metadsl import *
from metadsl_core import *
from .llvmlite_binding import *
from .llvmlite_ir import *
from .rules import register_ctypes


__all__ = ["CType", "CFunctionType", "make_c_wrapper", "compile_function"]


class CType(Expression):
    @expression
    @classmethod
    def c_int(cls) -> CType:
        ...

    @expression
    @classmethod
    def box(cls, tp: typing.Any) -> CType:
        ...


@register_ctypes
@rule
def ctype_cint() -> R[CType]:
    return CType.c_int(), lambda: CType.box(ctypes.c_int)


class CFunctionType(Expression):
    @expression
    @classmethod
    def create(cls, return_type: CType, *args: CType) -> CFunctionType:
        ...

    # Cannot represent type of C function
    @expression
    @classmethod
    def box(cls, tp: typing.Any) -> CFunctionType:
        ...

    @expression
    def __call__(self, ptr: int) -> typing.Callable:
        ...


@register_ctypes
@rule
def c_function_type_create_1(
    return_tp: typing.Any, arg1: typing.Any
) -> R[CFunctionType]:
    return (
        CFunctionType.create(CType.box(return_tp), CType.box(arg1)),
        lambda: CFunctionType.box(ctypes.CFUNCTYPE(return_tp, arg1)),
    )


@register_ctypes
@rule
def cfunc_call(fntype: typing.Any, ptr: int) -> R[typing.Callable]:
    return CFunctionType.box(fntype)(ptr), lambda: fntype(ptr)


@expression
def concat_strings(l: str, r: str) -> str:
    return l + r


register(default_rule(concat_strings))


@expression
def make_c_wrapper(
    module_builder: ModuleBuilder, original_fn_ref: FunctionReference
) -> Pair[ModuleBuilder, Function]:
    """
    Creates a new function that wraps the old function,
    making sure the calling convention is default.
    """
    module_builder, fn_ref = FunctionReference.create(
        module_builder,
        original_fn_ref.type,
        concat_strings("entry_", original_fn_ref.name),
    ).spread
    fn_builder = FunctionBuilder.create(fn_ref)
    fn_builder, block_ref = BlockReference.create("entry", fn_builder).spread
    block_builder = BlockBuilder.create(block_ref)
    block_builder, value = block_builder.call(
        original_fn_ref, fn_builder.arguments
    ).spread
    block_builder = block_builder.ret(value)
    block = Block.create(block_ref, block_builder)
    function = Function.create(fn_ref, Vec.create(block))
    return Pair.create(module_builder, function)


register(default_rule(make_c_wrapper))


@expression
def compile_function(
    module: Module,
    module_builder: ModuleBuilder,
    function_ref: FunctionReference,
    cfunctype: CFunctionType,
    optimization: int = 1,
) -> typing.Callable:
    module_builder, wrapper_fn = make_c_wrapper(module_builder, function_ref).spread
    module = Module.create(module.reference, module.functions.append(wrapper_fn))
    module_ref = ModuleRef.create(module.to_string())
    module_ref = module_ref.optimize(optimization)
    engine = ExecutionEngine.create(module_ref)
    return cfunctype(engine.get_function_address(wrapper_fn.reference.name))


register(default_rule(compile_function))



from metadsl import *
from metadsl_core import *

__all__ = [
    "register_llvmlite_ir_ref",
    "register_llvmlite_ir",
    "register_llvmlite_binding",
    "register_ctypes",
]

llvmlite_ir_ref_rules = RulesRepeatFold()
register_llvmlite_ir_ref = llvmlite_ir_ref_rules.append
rule_groups["llvmlite.ir (reference)"] = llvmlite_ir_ref_rules


llvmlite_ir_rules = RulesRepeatFold()
register_llvmlite_ir = llvmlite_ir_rules.append
rule_groups["llvmlite.ir"] = llvmlite_ir_rules


llvmlite_binding_rules = RulesRepeatFold()
register_llvmlite_binding = llvmlite_binding_rules.append
rule_groups["llvmlite.binding"] = llvmlite_binding_rules


ctypes_rules = RulesRepeatFold()
register_ctypes = ctypes_rules.append
rule_groups["ctypes"] = ctypes_rules



# import metadsl.python.expressions as py
# import metadsl.python.wraps as py_wraps
# from metadsl import *


# class TestInteger:
#     def test_from_int(self):
#         assert convert(py.Integer, 10) == py.Integer.from_int(10)

#     def test_add(self):
#         one = py.Integer.from_int(1)
#         assert py_wraps.Integer(one) + 10 == py_wraps.Integer(
#             one + py.Integer.from_int(10)
#         )


# class TestTuple:
#     def test_from_tuple(self):
#         assert convert(
#             py.Tuple[py.Integer], (0, 1)
#         ) == py.Tuple.from_items(
#             py.Integer, py.Integer.from_int(0), py.Integer.from_int(1)
#         )

#     def test_integer_tuple_getitem(self):
#         i = py.Tuple.from_items(
#             py.Integer, py.Integer.from_int(0), py.Integer.from_int(1)
#         )
#         assert py_wraps.IntegerTuple(i)[0] == py_wraps.Integer(
#             i[py.Integer.from_int(0)]
#         )


# class TestOptional:
#     def test_none(self):
#         assert convert(
#             py.Optional[py.Boolean], None
#         ) == py.Optional.none(py.Boolean)

#     def test_some(self):
#         assert convert(
#             py.Optional[py.Boolean], True
#         ) == py.Optional.some(py.Boolean.from_bool(True))


# class TestUnion:
#     def test_from_left(self):
#         assert convert(
#             py.Union[py.Integer, py.Boolean], 1
#         ) == py.Union.left(
#             py.Integer, py.Boolean, py.Integer.from_int(1)
#         )

#     def test_from_right(self):
#         assert convert(
#             py.Union[py.Integer, py.Boolean], True
#         ) == py.Union.right(
#             py.Integer, py.Boolean, py.Boolean.from_bool(True)
#         )



"""
Integeration with typez, to translate a "Rule" into a something that takes returns an iterable of
typez object we can use to display as we go.
"""

from typez import *
import metadsl
import typing
import inspect
import dataclasses
import typing_inspect
import warnings

__all__ = ["ExpressionDisplay", "SHOW_MODULE"]


@dataclasses.dataclass
class ExpressionDisplay:
    ref: metadsl.ExpressionReference
    typez_display: TypezDisplay = dataclasses.field(
        default_factory=lambda: TypezDisplay(Typez())
    )

    def __post_init__(self):
        ref = self.ref
        nodes = convert_to_nodes(ref)
        initial_node_id = str(ref.hash)
        self.typez_display.typez = Typez(
            states=States(initial=initial_node_id), nodes=nodes
        )

    def update(self, rule: str, label: typing.Optional[str] = None):
        new_nodes = convert_to_nodes(self.ref)
        self.typez_display.typez = dataclasses.replace(
            self.typez_display.typez,
            nodes={**(self.typez_display.typez.nodes or {}), **new_nodes},
            states=dataclasses.replace(
                self.typez_display.typez.states,
                states=[
                    *(
                        self.typez_display.typez.states.states or []
                        if self.typez_display.typez.states
                        else []
                    ),
                    State(node=str(self.ref.hash), rule=rule, label=label),
                ],
            ),
        )

    def _ipython_display_(self):
        self.typez_display._ipython_display_()


def convert_to_nodes(ref: metadsl.ExpressionReference) -> Nodes:
    """
    Converts an expression into a node mapping and also returns the ID for the top level node.
    """
    expressions = ref.expressions
    nodes: Nodes = {}
    for hash_, expr in expressions.expressions.items():
        node: typing.Union[CallNode, PrimitiveNode]
        value = expr.value
        if isinstance(value, metadsl.Expression):
            children = expr.children
            assert children
            func_str = function_or_type_repr(value.function)
            node = CallNode(
                type_params=typevars_to_typeparams(
                    metadsl.typing_tools.get_fn_typevars(value.function)
                )
                or None,
                function=f"{func_str}: {value._type_str}" if SHOW_TYPES else func_str,
                args=[str(a) for a in children.args] or None,
                kwargs={k: str(v) for k, v in children.kwargs.items()} or None,
            )
        else:
            node = PrimitiveNode(
                type=function_or_type_repr(type(value)), repr=str(value)
            )
        nodes[str(hash_)] = [str(expr.id), node]  # type: ignore
    return nodes


def typevars_to_typeparams(
    typevars: metadsl.typing_tools.TypeVarMapping
) -> typing.Dict[str, TypeInstance]:
    return {
        var.__name__: type_to_typeinstance(tp)  # type: ignore
        for var, tp in typevars.items()
    }


def type_to_typeinstance(tp: typing.Type) -> TypeInstance:
    """
    Converts a python type to the type instance, by folding through it
    """
    if issubclass(tp, metadsl.Expression):
        return DeclaredTypeInstance(
            type=function_or_type_repr(typing_inspect.get_origin(tp) or tp),
            params=typevars_to_typeparams(
                metadsl.typing_tools.match_types(
                    metadsl.typing_tools.get_origin_type(tp), tp
                )
            )
            or None,
        )
    return ExternalTypeInstance(repr=function_or_type_repr(tp))


_builtins = inspect.getmodule(int)


SHOW_MODULE = False

WARN_ON_TYPEVAR = False


def function_or_type_repr(o: typing.Union[typing.Type, typing.Callable]) -> str:
    """
    returns the name of the type or function prefixed by the module name if it isn't a builtin
    """
    if isinstance(o, typing.TypeVar):  # type: ignore
        if WARN_ON_TYPEVAR:
            warnings.warn(
                "Unresolved typevar in type argument, this means the typing is broken",
                RuntimeWarning,
            )
            return repr(o)
        else:
            raise RuntimeError("Unresolved typevar in type argument")

    # This is true for `typing.Any`
    if isinstance(o, typing._SpecialForm):
        return repr(o)

    name = o.__qualname__ or o.__name__
    # This is true for objects in typing
    if not name:
        return repr(o)
    if not SHOW_MODULE:
        return name
    module = inspect.getmodule(o)
    if module == _builtins:
        return name
    return f"{module.__name__}.{name}"



from __future__ import annotations

import pytest
import typing

from .expressions import *
from .normalized import *
from .expressions_test import TEST_EXPRESSIONS


@pytest.mark.parametrize("expr", TEST_EXPRESSIONS)
def test_expression_reference_identity(expr):
    ref = ExpressionReference.from_expression(expr)
    ref.verify_integrity()
    assert ref.normalized_expression.value == expr


@expression
def a(e: typing.Any) -> typing.Any:
    ...


@expression
def b(e: typing.Any) -> typing.Any:
    ...


@expression
def c() -> typing.Any:
    ...


@expression
def d() -> typing.Any:
    ...


def test_replace():
    ref = ExpressionReference.from_expression(a(b(c())))
    ref.verify_integrity()
    original_hashes, original_ids, original_expr = zip(
        *(
            (
                child.hash,
                child.normalized_expression.id,
                child.normalized_expression.value,
            )
            for child in ref.children
        )
    )

    assert original_expr == (a(b(c())), b(c()), c())

    ref.replace(a(b(d())))
    ref.verify_integrity()

    new_hashes, new_ids, new_expr = zip(
        *(
            (
                child.hash,
                child.normalized_expression.id,
                child.normalized_expression.value,
            )
            for child in ref.children
        )
    )

    assert new_expr == (a(b(d())), b(d()), d())

    # None of the hashes should be the same
    for i in (0, 1, 2):
        assert original_hashes[i] != new_hashes[i]

    # the first two ids should be the  same
    for i in (0, 1):
        assert original_ids[i] == new_ids[i]

    # the last should be different
    assert original_ids[2] != new_ids[2]


def test_replace_child():
    ref = ExpressionReference.from_expression(a(b(c())))

    ref.verify_integrity()

    original_hashes, original_ids, original_expr = zip(
        *(
            (
                child.hash,
                child.normalized_expression.id,
                child.normalized_expression.value,
            )
            for child in ref.children
        )
    )

    assert original_expr == (a(b(c())), b(c()), c())
    child_ref = list(ref.children)[-1]
    child_ref.replace(d())
    child_ref.verify_integrity()

    ref.verify_integrity()

    new_hashes, new_ids, new_expr = zip(
        *(
            (
                child.hash,
                child.normalized_expression.id,
                child.normalized_expression.value,
            )
            for child in ref.children
        )
    )

    assert new_expr == (a(b(d())), b(d()), d())

    # None of the hashes should be the same
    for i in (0, 1, 2):
        assert original_hashes[i] != new_hashes[i]

    # the first two ids should be the  same
    for i in (0, 1):
        assert original_ids[i] == new_ids[i]

    # the last should be different
    assert original_ids[2] != new_ids[2]


@expression
def e(a: typing.Any, b: typing.Any) -> typing.Any:
    ...


@expression
def f(e: typing.Any) -> typing.Any:
    ...


@expression
def g(e: typing.Any) -> typing.Any:
    ...


def test_graph_subtree():
    """
    Verify processed in correct order
    """
    orig = e(f(g(d())), d())
    ref = ExpressionReference.from_expression(orig)
    ref.verify_integrity()
    assert ref.normalized_expression.value == orig


def test_replace_subtree_same_id():
    orig_expr = e(d(), d())
    new_expr = e(c(), c())

    ref = ExpressionReference.from_expression(e(orig_expr, d()))
    for child in ref.children:
        if child.normalized_expression.value == orig_expr:
            orig_id = child.normalized_expression.id
            break

    ref.replace(e(new_expr, d()))

    for child in ref.children:
        if child.normalized_expression.value == new_expr:
            new_id = child.normalized_expression.id
            break

    assert orig_id == new_id


def test_doesnt_remember_replacements():
    ref = ExpressionReference.from_expression(a(b(c())))
    ref.verify_integrity()
    ref.replace(a(b(d())))
    ref.verify_integrity()
    ref.replace(a(b(c())))
    ref.verify_integrity()

    assert ref.normalized_expression.value == a(b(c()))



from __future__ import annotations

# TODO: make supertype things dunder so they don't interfere
# Support level of indirection to allow for mutable functions
# (support this as Mutates[...] like fortran, in, out, inout)
# Meta level transformations... Taking things like returning
# pairs of things and changing to mutate passed in args!
# Basically rewrites at the meta level? Or too many layers?
# Tradeoff  of making it harder to have client implementation with
# confusion
# CONCLUSION: Dont do at meta level, just support first class
# with Mutates[...] added to args which creates GDT of the arg names
# + the return type, if there is one.

# TODO: Allow doing *(CreateThing(x) for x in ...)
# To match a list of a certain expression

# TODO: Support returning multiple values?: NAW!


# TODO: Add dataclass like operator for storing  GDTs, so you don't have to define
# each getter method...
# To do this, need to be able to override __init__?

# How do I get access to original init? With super?
# Can't subclass expression... So instead wrap dataclass class in decorator.

# TODO: Why are docstrings wrong for Expressions?
# A: Ah.... I am trying to do init instead of create

# TODO: Can we have annotations be lazy in some way so we can define type below and then have it added?
import typing

import llvmlite.ir as ir

from metadsl import *
from metadsl_core import *

from .rules import register_llvmlite_ir_ref, register_llvmlite_ir


__all__ = [
    "ModuleReference",
    "ModuleBuilder",
    "Type",
    "FunctionType",
    "FunctionReference",
    "FunctionBuilder",
    "BlockReference",
    "BlockBuilder",
    "Value",
    "Block",
    "Function",
    "Module",
    "Value",
]


@hash_value.register
def hash_llvmlite_type(value: ir.Type) -> int:
    return hash((type(value), value._to_string()))


class ModuleReference(Expression):
    @expression
    @classmethod
    def create(cls, name: str) -> ModuleReference:
        ...

    @expression
    @classmethod
    def box(cls, module: ir.Module) -> ModuleReference:
        ...


class ModuleBuilder(Expression):
    @expression
    @classmethod
    def create(cls, mod_ref: ModuleReference) -> ModuleBuilder:
        ...

    @expression
    @classmethod
    def box(cls, module: ir.Module) -> ModuleBuilder:
        ...


class Type(Expression):
    @expression
    @classmethod
    def create_int(cls, bits: int) -> Type:
        ...

    @expression
    @classmethod
    def box(cls, type: ir.Type) -> Type:
        ...


class FunctionType(Expression):
    @expression
    @classmethod
    def create(cls, return_type: Type, *arg_types: Type) -> FunctionType:
        ...

    @expression
    @classmethod
    def box(cls, type: ir.FunctionType) -> FunctionType:
        ...


class FunctionReference(Expression):
    @expression
    @classmethod
    def create(
        cls,
        mod_builder: ModuleBuilder,
        type: FunctionType,
        name: str,
        calling_convention: str = "",
    ) -> Pair[ModuleBuilder, FunctionReference]:
        ...

    @expression
    @classmethod
    def box(cls, function: ir.Function) -> FunctionReference:
        ...

    @expression  # type: ignore
    @property
    def type(self) -> FunctionType:
        ...

    @expression  # type: ignore
    @property
    def name(self) -> str:
        ...


class FunctionBuilder(Expression):
    @expression
    @classmethod
    def create(cls, function_ref: FunctionReference) -> FunctionBuilder:
        ...

    @expression
    @classmethod
    def box(cls, function: ir.Function) -> FunctionBuilder:
        ...

    @expression  # type: ignore
    @property
    def arguments(self) -> Vec[Value]:
        ...


class BlockReference(Expression):
    @expression
    @classmethod
    def create(
        cls, name: str, fn_builder: FunctionBuilder
    ) -> Pair[FunctionBuilder, BlockReference]:
        ...

    @expression
    @classmethod
    def box(cls, block: ir.Block) -> BlockReference:
        ...


class BlockBuilder(Expression):
    @expression
    @classmethod
    def create(cls, block: BlockReference) -> BlockBuilder:
        ...

    @expression
    @classmethod
    def box(cls, builder: ir.IRBuilder) -> BlockBuilder:
        ...

    @expression
    def icmp_signed(
        self, operator: str, left: Value, right: Value
    ) -> Pair[BlockBuilder, Value]:
        ...

    @expression
    def ret(self, value: Value) -> BlockBuilder:
        ...

    @expression
    def call(
        self, function: FunctionReference, args: Vec[Value]
    ) -> Pair[BlockBuilder, Value]:
        ...

    @expression
    def cbranch(
        self, condition: Value, true: BlockReference, false: BlockReference
    ) -> BlockBuilder:
        ...

    @expression
    def add(self, left: Value, right: Value) -> Pair[BlockBuilder, Value]:
        ...

    @expression
    def sub(self, left: Value, right: Value) -> Pair[BlockBuilder, Value]:
        ...


class Value(Expression):
    @expression
    @classmethod
    def constant(cls, type: Type, value: typing.Any) -> Value:
        ...

    @expression
    @classmethod
    def box(cls, value: ir.Value) -> Value:
        ...


class Block(Expression):
    @expression
    @classmethod
    def create(cls, block_ref: BlockReference, block_builder: BlockBuilder) -> Block:
        ...

    @expression
    @classmethod
    def box(cls, block: ir.Block) -> Block:
        ...


class Function(Expression):
    @expression
    @classmethod
    def create(cls, reference: FunctionReference, blocks: Vec[Block]) -> Function:
        ...

    @expression  # type: ignore
    @property
    def reference(self) -> FunctionReference:
        ...

    @expression  # type: ignore
    @property
    def blocks(self) -> Vec[Block]:
        ...

    @expression
    @classmethod
    def box(cls, function: ir.Function) -> Function:
        ...


class Module(Expression):
    @expression
    @classmethod
    def create(cls, module_ref: ModuleReference, functions: Vec[Function]) -> Module:
        ...

    @expression
    @classmethod
    def box(cls, module: ir.Module) -> Module:
        ...

    @expression  # type: ignore
    @property
    def reference(self) -> ModuleReference:
        ...

    @expression  # type: ignore
    @property
    def functions(self) -> Vec[Function]:
        ...

    @expression
    def to_string(self) -> str:
        ...


@register_llvmlite_ir_ref
@rule
def module_reference_box(name: str) -> R[ModuleReference]:
    return ModuleReference.create(name), lambda: ModuleReference.box(ir.Module(name))


@register_llvmlite_ir_ref
@rule
def module_builder_box(mod: ir.Module) -> R[ModuleBuilder]:
    return (ModuleBuilder.create(ModuleReference.box(mod)), ModuleBuilder.box(mod))


@register_llvmlite_ir_ref
@rule
def type_create_int_box(bits: int) -> R[Type]:
    return Type.create_int(bits), lambda: Type.box(ir.IntType(bits))


@register_llvmlite_ir_ref
@rule
def function_type_box_1(ret: ir.Type, arg1: ir.Type) -> R[FunctionType]:
    return (
        FunctionType.create(Type.box(ret), Type.box(arg1)),
        lambda: FunctionType.box(ir.FunctionType(ret, (arg1,))),
    )


@register_llvmlite_ir_ref
@rule
def function_type_box_2(ret: ir.Type, arg1: ir.Type, arg2: ir.Type) -> R[FunctionType]:
    return (
        FunctionType.create(Type.box(ret), Type.box(arg1), Type.box(arg2)),
        lambda: FunctionType.box(ir.FunctionType(ret, (arg1, arg2))),
    )


@register_llvmlite_ir_ref
@rule
def function_type_box_3(
    ret: ir.Type, arg1: ir.Type, arg2: ir.Type, arg3: ir.Type
) -> R[FunctionType]:
    return (
        FunctionType.create(
            Type.box(ret), Type.box(arg1), Type.box(arg2), Type.box(arg3)
        ),
        lambda: FunctionType.box(ir.FunctionType(ret, (arg1, arg2, arg3))),
    )


@register
@rule
def function_reference_type(
    module: ModuleBuilder, type: FunctionType, name: str, calling_convention: str
) -> R[FunctionType]:
    return (
        FunctionReference.create(module, type, name, calling_convention).right.type,
        type,
    )


@register
@rule
def function_reference_name(
    module: ModuleBuilder, type: FunctionType, name: str, calling_convention: str
) -> R[str]:
    return (
        FunctionReference.create(module, type, name, calling_convention).right.name,
        name,
    )


@register_llvmlite_ir_ref
@rule
def function_reference_create_builder(
    mod: ir.Module, function_type: FunctionType, name: str, calling_convention: str
) -> R[ModuleBuilder]:
    return (
        FunctionReference.create(
            ModuleBuilder.box(mod), function_type, name, calling_convention
        ).left,
        ModuleBuilder.box(mod),
    )


@register_llvmlite_ir_ref
@rule
def function_reference_box(
    mod: ir.Module, fntype: ir.FunctionType, name: str, calling_convention: str
) -> R[FunctionReference]:
    def inner() -> FunctionReference:
        fn = ir.Function(mod, fntype, name)
        fn.calling_convention = calling_convention
        return FunctionReference.box(fn)

    return (
        FunctionReference.create(
            ModuleBuilder.box(mod), FunctionType.box(fntype), name, calling_convention
        ).right,
        inner,
    )


@register_llvmlite_ir_ref
@rule
def function_builder_box(fn: ir.Function,) -> R[FunctionBuilder]:
    return (
        FunctionBuilder.create(FunctionReference.box(fn)),
        lambda: FunctionBuilder.box(fn),
    )


@register_llvmlite_ir_ref
@rule
def function_builder_arguments(function: ir.Function) -> R[Vec[Value]]:
    return (
        FunctionBuilder.box(function).arguments,
        lambda: Vec.create(*(Value.box(a) for a in function.args)),
    )


@register_llvmlite_ir_ref
@rule
def block_reference_box(
    function: ir.Function, name: str
) -> R[Pair[FunctionBuilder, BlockReference]]:
    return (
        BlockReference.create(name, FunctionBuilder.box(function)),
        lambda: Pair.create(
            FunctionBuilder.box(function),
            BlockReference.box(function.append_basic_block(name)),
        ),
    )


@register_llvmlite_ir_ref
@rule
def block_builder_box(block: ir.Block) -> R[BlockBuilder]:
    return (
        BlockBuilder.create(BlockReference.box(block)),
        lambda: BlockBuilder.box(ir.IRBuilder(block)),
    )


@register_llvmlite_ir_ref
@rule
def value_constant(tp: ir.Type, value: typing.Any) -> R[Value]:
    return (
        Value.constant(Type.box(tp), value),
        lambda: Value.box(ir.Constant(tp, value)),
    )


@register_llvmlite_ir_ref
@rule
def builder_icmp_unsigned(
    builder: ir.IRBuilder, operator: str, left: ir.Value, right: ir.Value
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.icmp_signed(operator, Value.box(left), Value.box(right)),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.icmp_unsigned(operator, left, right))
        ),
    )


@register_llvmlite_ir_ref
@rule
def builder_ret(builder: ir.IRBuilder, value: ir.Value) -> R[BlockBuilder]:
    builder_ = BlockBuilder.box(builder)

    def inner():
        builder.ret(value)
        return builder_

    return (builder_.ret(Value.box(value)), inner)


@register_llvmlite_ir_ref
@rule
def builder_call_1(
    builder: ir.IRBuilder, function: ir.Function, arg1: ir.Value
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.call(FunctionReference.box(function), Vec.create(Value.box(arg1))),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.call(function, (arg1,)))
        ),
    )


@register_llvmlite_ir_ref
@rule
def builder_call_2(
    builder: ir.IRBuilder, function: ir.Function, arg1: ir.Value, arg2: ir.Value
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.call(
            FunctionReference.box(function),
            Vec.create(Value.box(arg1), Value.box(arg2)),
        ),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.call(function, (arg1, arg2)))
        ),
    )


@register_llvmlite_ir_ref
@rule
def builder_call_3(
    builder: ir.IRBuilder,
    function: ir.Function,
    arg1: ir.Value,
    arg2: ir.Value,
    arg3: ir.Value,
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.call(
            FunctionReference.box(function),
            Vec.create(Value.box(arg1), Value.box(arg2), Value.box(arg3)),
        ),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.call(function, (arg1, arg2, arg3)))
        ),
    )


# # TODO: Support creating iterations of values in matches to make this variadic:
# """
# @register_llvmlite_ir_ref
# @rule
# def builder_call(
#     builder: ir.IRBuilder, function: ir.Function, values: typing.Sequence[ir.Value]
# ) -> R[Pair[BlockBuilder, Value]]:
#     builder_ = BlockBuilder.box(builder)
#     return (
#         builder_.call(
#             FunctionReference.box(function),
#             Vec.create(*(Value.box(value) for value in values)),
#         ),
#         lambda: Pair[BlockBuilder, Value].create(
#             builder_, Value.box(builder.call(function, tuple(values)))
#         ),
#     )
# """


@register_llvmlite_ir_ref
@rule
def builder_cbranch(
    builder: ir.IRBuilder, cond: ir.Value, true: ir.Block, false: ir.Block
) -> R[BlockBuilder]:
    builder_ = BlockBuilder.box(builder)

    def inner() -> BlockBuilder:
        builder.cbranch(cond, true, false)
        return builder_

    return (
        builder_.cbranch(
            Value.box(cond), BlockReference.box(true), BlockReference.box(false)
        ),
        inner,
    )


@register_llvmlite_ir_ref
@rule
def builder_add(
    builder: ir.IRBuilder, left: ir.Value, right: ir.Value
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.add(Value.box(left), Value.box(right)),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.add(left, right))
        ),
    )


@register_llvmlite_ir_ref
@rule
def builder_sub(
    builder: ir.IRBuilder, left: ir.Value, right: ir.Value
) -> R[Pair[BlockBuilder, Value]]:
    builder_ = BlockBuilder.box(builder)
    return (
        builder_.sub(Value.box(left), Value.box(right)),
        lambda: Pair[BlockBuilder, Value].create(
            builder_, Value.box(builder.sub(left, right))
        ),
    )


@register_llvmlite_ir
@rule
def block_box(block: ir.Block, builder: ir.IRBuilder) -> R[Block]:
    return (
        Block.create(BlockReference.box(block), BlockBuilder.box(builder)),
        lambda: Block.box(block),
    )


@register_llvmlite_ir
@rule
def function_box_1(fn: ir.Function, block1: ir.Block) -> R[Function]:
    return (
        Function.create(FunctionReference.box(fn), Vec.create(Block.box(block1))),
        lambda: Function.box(fn),
    )


@register_llvmlite_ir
@rule
def function_box_2(fn: ir.Function, block1: ir.Block, block2: ir.Block) -> R[Function]:
    return (
        Function.create(
            FunctionReference.box(fn), Vec.create(Block.box(block1), Block.box(block2))
        ),
        lambda: Function.box(fn),
    )


@register_llvmlite_ir
@rule
def function_box_3(
    fn: ir.Function, block1: ir.Block, block2: ir.Block, block3: ir.Block
) -> R[Function]:
    return (
        Function.create(
            FunctionReference.box(fn),
            Vec.create(Block.box(block1), Block.box(block2), Block.box(block3)),
        ),
        lambda: Function.box(fn),
    )


@register_llvmlite_ir
@rule
def function_box_4(
    fn: ir.Function,
    block1: ir.Block,
    block2: ir.Block,
    block3: ir.Block,
    block4: ir.Block,
) -> R[Function]:
    return (
        Function.create(
            FunctionReference.box(fn),
            Vec.create(
                Block.box(block1),
                Block.box(block2),
                Block.box(block3),
                Block.box(block4),
            ),
        ),
        lambda: Function.box(fn),
    )


@register_llvmlite_ir
@rule
def function_box_5(
    fn: ir.Function,
    block1: ir.Block,
    block2: ir.Block,
    block3: ir.Block,
    block4: ir.Block,
    block5: ir.Block,
) -> R[Function]:
    return (
        Function.create(
            FunctionReference.box(fn),
            Vec.create(
                Block.box(block1),
                Block.box(block2),
                Block.box(block3),
                Block.box(block4),
                Block.box(block5),
            ),
        ),
        lambda: Function.box(fn),
    )


@register
@rule
def function_reference(
    fn: FunctionReference, blocks: Vec[Block]
) -> R[FunctionReference]:
    return Function.create(fn, blocks).reference, fn


@register
@rule
def function_blocks(fn: FunctionReference, blocks: Vec[Block]) -> R[Vec[Block]]:
    return Function.create(fn, blocks).blocks, blocks


@register_llvmlite_ir
@rule
def module_box_1(mod: ir.Module, fn1: ir.Function) -> R[Module]:
    return (
        Module.create(ModuleReference.box(mod), Vec.create(Function.box(fn1))),
        Module.box(mod),
    )


@register_llvmlite_ir
@rule
def module_box_2(mod: ir.Module, fn1: ir.Function, fn2: ir.Function) -> R[Module]:
    return (
        Module.create(
            ModuleReference.box(mod), Vec.create(Function.box(fn1), Function.box(fn2))
        ),
        Module.box(mod),
    )


@register_llvmlite_ir
@rule
def module_box_3(mod: ir.Module, fn1: ir.Function, fn2: ir.Function, fn3: ir.Function) -> R[Module]:
    return (
        Module.create(
            ModuleReference.box(mod), Vec.create(Function.box(fn1), Function.box(fn2), Function.box(fn3))
        ),
        Module.box(mod),
    )


@register_llvmlite_ir
@rule
def module_to_string(mod: ir.Module) -> R[str]:
    return Module.box(mod).to_string(), lambda: str(mod)


@register
@rule
def module_reference(
    reference: ModuleReference, functions: Vec[Function]
) -> R[ModuleReference]:
    return Module.create(reference, functions).reference, reference


@register
@rule
def module_functions(
    reference: ModuleReference, functions: Vec[Function]
) -> R[Vec[Function]]:
    return Module.create(reference, functions).functions, functions




# from __future__ import annotations


# from metadsl import *
# from metadsl.python.expressions import *

# __all__ = [
#     "NDArray",
#     "DType",
#     "arange",
# ]

# class NDArray(Expression):
#     @expression
#     def shape(self) -> Tuple[Integer]:
#         ...

#     @expression
#     def __add__(self, other: NDArray) -> NDArray:
#         ...

#     @expression
#     def __mul__(self, other: NDArray) -> NDArray:
#         ...

#     @expression
#     def __getitem__(self, idxs: Union[Integer, Tuple[Integer]]) -> NDArray:
#         ...

#     # @staticmethod
#     # @expression
#     # def __array_ufunc__(
#     #     ufunc: "UFunc", method: "UFuncMethod", kwargs: "UFuncKwargs", *inputs: NDArray
#     # ) -> NDArray:
#     #     ...

#     # @staticmethod
#     # @expression
#     # def __array_function__(
#     #     func: NumPyFunction, args: Tuple[Instance], kwargs: Instance
#     # ) -> NDArray:
#     #     pass


# # class NumPyFunction(Instance):
# #     pass


# class DType(Expression):
#     pass


# # class UFunc(Instance):
# #     pass


# # class UFuncKwargs(Instance):
# #     pass


# # class UFuncMethod(Instance):
# #     """
# #     https://docs.scipy.org/doc/numpy-1.13.0/neps/ufunc-overrides.html#proposed-interface

# #     method is a string indicating how the Ufunc was called, either "__call__" to indicate it was called directly, or one of its methods: "reduce", "accumulate", "reduceat", "outer", or "at".
# #     """

# #     @call(lambda self, k, a, c, f: k)
# #     def match(
# #         self, call: T, reduce: T, accumulate: T, reduceat: T, outer: T, at: T
# #     ) -> T:
# #         ...


# # class ArrayOrder(Expression):
# #     @property  # type: ignore
# #     @staticmethod
# #     @expression
# #     def k() -> ArrayOrder:
# #         ...

# #     @property  # type: ignore
# #     @staticmethod
# #     @expression
# #     def a() -> ArrayOrder:
# #         ...

# #     @property  # type: ignore
# #     @staticmethod
# #     @expression
# #     def c() -> ArrayOrder:
# #         ...

# #     @property  # type: ignore
# #     @staticmethod
# #     @expression
# #     def f() -> ArrayOrder:
# #         ...


# # @expression
# # def array(
# #     object: NDArray,
# #     dtype: Optional[DType],
# #     copy: Boolean,
# #     order: Optional[ArrayOrder],
# #     subok: Optional[Boolean],
# #     ndmin: Optional[Integer],
# # ) -> NDArray:
# #     ...


# @expression
# def arange(
#     start: Optional[Number],
#     stop: Number,
#     step: Optional[Number],
#     dtype: Optional[DType],
# ) -> NDArray:
#     ...



# from __future__ import annotations

# import numbers
# import typing

# import metadsl.python.expressions as py
# import typing_inspect
# from metadsl import *
# from metadsl.typing_tools import safe_issubclass

# __all__ = ["Integer", "Number", "Number", "IntegerTuple"]

# T = typing.TypeVar("T")


# @register_converter
# def _bool_to_boolean(t: typing.Type[T], value: object) -> T:
#     if safe_issubclass(t, py.Boolean) and isinstance(value, bool):
#         return typing.cast(T, py.Boolean.from_bool(value))
#     return NotImplemented


# class Integer(Wrap[py.Integer]):
#     @wrap(py.Integer.__add__)
#     def __add__(self, other: object) -> Integer:
#         ...

#     @wrap(py.Integer.__mul__)
#     def __mul__(self, other: object) -> Integer:
#         ...


# @register_converter
# def _int_to_integer(t: typing.Type[T], value: object) -> T:
#     if (
#         safe_issubclass(t, py.Integer)
#         and isinstance(value, int)
#         and not isinstance(value, bool)
#     ):
#         return typing.cast(T, py.Integer.from_int(value))
#     return NotImplemented


# class Number(Wrap[py.Number]):
#     @wrap(py.Number.__add__)
#     def __add__(self, other: object) -> Number:
#         ...

#     @wrap(py.Number.__mul__)
#     def __mul__(self, other: object) -> Number:
#         ...


# @register_converter
# def _number_to_number(t: typing.Type[T], value: object) -> T:
#     if safe_issubclass(t, py.Number) and isinstance(value, numbers.Number):
#         return typing.cast(T, py.Number.from_number(value))
#     return NotImplemented


# class IntegerTuple(Wrap[py.Tuple[py.Integer]]):
#     @wrap(py.Tuple.__getitem__)
#     def __getitem__(self, index: object) -> Integer:
#         ...


# @register_converter
# def _tuple_to_tuple(t: typing.Type[T], value: object) -> T:
#     if safe_issubclass(t, py.Tuple) and isinstance(value, tuple):
#         inner_type, = typing_inspect.get_args(t)
#         return typing.cast(
#             T,
#             py.Tuple.from_items(
#                 inner_type, *(convert(inner_type, a) for a in value)
#             ),
#         )
#     return NotImplemented


# @register_converter
# def _to_optional(t: typing.Type[T], value: object) -> T:
#     if safe_issubclass(t, py.Optional):
#         inner_type, = typing_inspect.get_args(t)
#         if value is None:
#             return typing.cast(T, py.Optional.none(inner_type))
#         return typing.cast(T, py.Optional.some(convert(inner_type, value)))
#     return NotImplemented


# @register_converter
# def _to_union(t: typing.Type[T], value: object) -> T:
#     if safe_issubclass(t, py.Union):
#         left_type, right_type = typing_inspect.get_args(t)

#         try:
#             left_converted = convert(left_type, value)
#         except NotImplementedError:
#             left_converted = None

#         try:
#             right_converted = convert(right_type, value)
#         except NotImplementedError:
#             right_converted = None
#         if left_converted is not None and right_converted is not None:
#             raise ValueError(f"Ambigious conversion: {value} to {t}")

#         if left_converted is None and right_converted is None:
#             return NotImplemented
#         if left_converted:
#             return typing.cast(
#                 T, py.Union.left(left_type, right_type, left_converted)
#             )
#         return typing.cast(
#             T, py.Union.right(left_type, right_type, right_converted)
#         )
#     return NotImplemented
